@@ -31,9 +31,9 @@ import typing
 import logging
 import functools
 import itertools
+import contextlib
 import contextvars
 import collections
-from contextlib import suppress
 from pathlib import Path
 from pprint import pformat
 
@@ -44,10 +44,13 @@ from grill.tokens import ids
 
 logger = logging.getLogger(__name__)
 
-Repository = repo = contextvars.ContextVar('Repository')
+Repository = contextvars.ContextVar('Repository')
 
-_PRIM_GRILL_KEY = 'grill'
-_PRIM_FIELDS_KEY = 'fields'
+_TAXA_KEY = 'taxa'
+_FIELDS_KEY = 'fields'
+_ASSETINFO_KEY = 'grill'
+_ASSETINFO_TAXA_KEY = f'{_ASSETINFO_KEY}:{_TAXA_KEY}'
+_ASSETINFO_FIELDS_KEY = f'{_ASSETINFO_KEY}:{_FIELDS_KEY}'
 
 # Taxonomy rank handles the grill classification and grouping of assets.
 _TAXONOMY_NAME = 'Taxonomy'
@@ -55,6 +58,13 @@ _TAXONOMY_ROOT_PATH = Sdf.Path.absoluteRootPath.AppendChild(_TAXONOMY_NAME)
 _TAXONOMY_UNIQUE_ID = ids.CGAsset.cluster  # High level organization of our assets.
 _TAXONOMY_FIELDS = types.MappingProxyType({_TAXONOMY_UNIQUE_ID.name: _TAXONOMY_NAME})
 _UNIT_UNIQUE_ID = ids.CGAsset.item  # Entry point for meaningful composed assets.
+_UNIT_ORIGIN_PATH = Sdf.Path.absoluteRootPath.AppendChild("Origin")
+
+# Composition filters for asset edits
+_ASSET_UNIT_QUERY_FILTER = Usd.PrimCompositionQuery.Filter()
+_ASSET_UNIT_QUERY_FILTER.dependencyTypeFilter = Usd.PrimCompositionQuery.DependencyTypeFilter.Direct
+_ASSET_UNIT_QUERY_FILTER.hasSpecsFilter = Usd.PrimCompositionQuery.HasSpecsFilter.HasSpecs
+_ASSET_UNIT_QUERY_FILTER.arcTypeFilter = Usd.PrimCompositionQuery.ArcTypeFilter.ReferenceOrPayload
 
 
 @functools.lru_cache(maxsize=None)
@@ -66,18 +76,17 @@ def fetch_stage(identifier: str) -> Usd.Stage:
     If an open matching `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_ is found on the `global cache <https://graphics.pixar.com/usd/docs/api/class_usd_utils_stage_cache.html>`_, return it.
     Otherwise open it, populate the `cache <https://graphics.pixar.com/usd/docs/api/class_usd_utils_stage_cache.html>`_ and return it.
     """
-    rootf = UsdAsset(identifier)
+    layer_id = UsdAsset(identifier).name
     cache = UsdUtils.StageCache.Get()
-    repo_path = repo.get()
+    repo_path = Repository.get()
     resolver_ctx = Ar.DefaultResolverContext([str(repo_path)])
     with Ar.ResolverContextBinder(resolver_ctx):
-        layer_id = rootf.name
         logger.debug(f"Searching for {layer_id}")
         layer = Sdf.Layer.Find(layer_id)
         if not layer:
             logger.debug(f"Layer {layer_id} was not found open. Attempting to open it.")
             if not Sdf.Layer.FindOrOpen(layer_id):
-                logger.debug(f"Layer {layer_id} does not exist on repository path: {resolver_ctx.GetSearchPath()}. Creating a new one.")
+                logger.debug(f"Layer {layer_id} does not exist on repository path: {repo_path}. Creating a new one.")
                 # we first create a layer under our repo
                 tmp_new_layer = Sdf.Layer.CreateNew(str(repo_path / layer_id))
                 # delete it since it will have an identifier with the full path,
@@ -87,7 +96,6 @@ def fetch_stage(identifier: str) -> Usd.Stage:
                 #   In the meantime, we need to create the layer first on disk.
                 del tmp_new_layer
             stage = Usd.Stage.Open(layer_id)
-            logger.debug(f"Root layer: {stage.GetRootLayer()}")
             logger.debug(f"Opened stage: {stage}")
             cache_id = cache.Insert(stage)
             logger.debug(f"Added stage for {layer_id} with cache ID: {cache_id.ToString()}.")
@@ -95,7 +103,7 @@ def fetch_stage(identifier: str) -> Usd.Stage:
             logger.debug(f"Layer was open. Found: {layer}")
             stage = cache.FindOneMatching(layer)
             if not stage:
-                logger.debug(f"Could not find stage on the cache.")
+                logger.debug("Could not find stage on the cache.")
                 stage = Usd.Stage.Open(layer)
                 cache_id = cache.Insert(stage)
                 logger.debug(f"Added stage for {layer} with cache ID: {cache_id.ToString()}.")
@@ -136,66 +144,103 @@ def define_taxon(stage: Usd.Stage, name: str, *, references: tuple.Tuple[Usd.Pri
         raise ValueError(f"Got invalid id_field keys: {', '.join(invalid_fields)}. Allowed: {', '.join(ids.CGAsset.__members__)}")
 
     with taxonomy_context(stage):
-        prim = stage.CreateClassPrim(_TAXONOMY_ROOT_PATH.AppendChild(name))
+        prim = stage.DefinePrim(_TAXONOMY_ROOT_PATH.AppendChild(name))
         for reference in references:
             prim.GetReferences().AddInternalReference(reference.GetPath())
-        prim.CreateAttribute("label", Sdf.ValueTypeNames.String)
+        prim.CreateAttribute("label", Sdf.ValueTypeNames.String, custom=False)
         taxon_fields = {**fields, _TAXONOMY_UNIQUE_ID.name: name}
-        prim.SetCustomDataByKey(_PRIM_GRILL_KEY, {_PRIM_FIELDS_KEY: taxon_fields, "taxa": {name: 0}})
+        prim.SetAssetInfoByKey(_ASSETINFO_KEY, {_FIELDS_KEY: taxon_fields, _TAXA_KEY: {name: 0}})
 
     return prim
 
 
+def _iter_taxa(stage, taxon1, *taxonN, predicate=Usd.PrimAllPrimsPredicate):
+    """Iterate over prims that inherit from the given taxa."""
+    it = iter(Usd.PrimRange.Stage(stage, predicate=predicate))
+    taxa_names = {i if isinstance(i, str) else i.GetName() for i in (taxon1, *taxonN)}
+    for prim in it:
+        if prim.GetPath().HasPrefix(_TAXONOMY_ROOT_PATH):
+            # Ignore prims from the taxonomy hierarchy as they're not
+            # taxa members but the definition themselves.
+            it.PruneChildren()
+        elif taxa_names.intersection(prim.GetAssetInfoByKey(_ASSETINFO_TAXA_KEY) or {}):
+            yield prim
+
+
+def create_many(taxon, names, labels=tuple()) -> typing.List[Usd.Prim]:
+    """Create a new taxon member for each of the provided names.
+
+    a unit member of the given ``taxon``, with an optional display label.
+
+    When creating hundreds of thousands of members, this provides a slight performance improvement (around 15% on average) over :func:`create`.
+
+    The new members will be created as `prims <https://graphics.pixar.com/usd/docs/api/class_usd_prim.html>`_ on the given ``taxon``'s `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_.
+
+    .. seealso:: :func:`define_taxon` :func:`create`
+    """
+    stage = taxon.GetStage()
+    taxon_path = taxon.GetPath()
+    taxon_fields = _get_id_fields(taxon)
+
+    current_asset_name, root_layer = _root_asset(stage)
+    new_asset_name = UsdAsset(current_asset_name.get(**taxon_fields))
+    # Edits will go to the first layer that matches a valid pipeline identifier
+    # TODO: Evaluate if this agreement is robust enough for different workflows.
+
+    # Some workflows like houdini might load layers without permissions to edit
+    # Since we know we are in a valid pipeline layer, temporarily allow edits
+    # for our operations, then restore original permissions.
+    current_permission = root_layer.permissionToEdit
+    root_layer.SetPermissionToEdit(True)
+
+    # existing = {i.GetName() for i in _iter_taxa(taxon.GetStage(), *taxon.GetCustomDataByKey(_ASSETINFO_TAXA_KEY))}
+    taxonomy_layer = _find_layer_matching(_TAXONOMY_FIELDS, stage.GetLayerStack())
+    taxonomy_id = str(Path(taxonomy_layer.realPath).relative_to(Repository.get()))
+
+    scope_path = stage.GetPseudoRoot().GetPath().AppendPath(taxon.GetName())
+    scope = stage.GetPrimAtPath(scope_path)
+
+    def _create(name, label):
+        path = scope_path.AppendChild(name)
+        prim = stage.GetPrimAtPath(path)
+        if prim:
+            return prim
+        assetid = new_asset_name.get(**{_UNIT_UNIQUE_ID.name: name})
+        asset_stage = fetch_stage(assetid)
+        asset_layer = asset_stage.GetRootLayer()
+        asset_layer.subLayerPaths.append(taxonomy_id)
+        asset_origin = asset_stage.DefinePrim(_UNIT_ORIGIN_PATH)
+        asset_origin.GetInherits().AddInherit(taxon_path)
+        asset_stage.SetDefaultPrim(asset_origin)
+        if label:
+            label_attr = asset_origin.GetAttribute("label")
+            label_attr.Set(label)
+
+        over_prim = stage.OverridePrim(path)
+        over_prim.GetReferences().AddReference(asset_layer.identifier)
+        return over_prim
+
+    labels = itertools.chain(labels, itertools.repeat(""))
+    with Usd.EditContext(stage, root_layer):
+        # Scope collecting all assets of the same type
+        if not scope:
+            scope = stage.DefinePrim(scope_path)
+        if not scope.IsModel():
+            Usd.ModelAPI(scope).SetKind(Kind.Tokens.assembly)
+        prims = [_create(name, label) for name, label in zip(names, labels)]
+
+    root_layer.SetPermissionToEdit(current_permission)
+    return prims
+
+
 def create(taxon: Usd.Prim, name: str, label: str = "") -> Usd.Prim:
     """Create a unit member of the given ``taxon``, with an optional display label.
 
     The new member will be created as a `prim <https://graphics.pixar.com/usd/docs/api/class_usd_prim.html>`_ on the given ``taxon``'s `stage <https://graphics.pixar.com/usd/docs/api/class_usd_stage.html>`_.
 
-    .. seealso:: :func:`define_taxon`
+    .. seealso:: :func:`define_taxon` and :func:`create_many`
     """
-    stage = taxon.GetStage()
-    new_tokens = {**_get_id_fields(taxon), _UNIT_UNIQUE_ID.name: name}
-
-    # Edits will go to the first layer that matches a valid pipeline identifier
-    # TODO: Evaluate if this agreement is robust enough for different workflows.
-    current_asset_name, root_layer = _root_asset(stage)
-    # Some workflows like houdini might load layers without permissions to edit
-    # Since we know we are in a valid pipeline layer, temporarily allow edits
-    # for our operations, then restore original permissions.
-    current_permission = root_layer.permissionToEdit
-    root_layer.SetPermissionToEdit(True)
-
-    new_asset_name = current_asset_name.get(**new_tokens)
-
-    with Usd.EditContext(stage, root_layer):
-        # Scope collecting all assets of the same type
-        scope_path = stage.GetPseudoRoot().GetPath().AppendPath(taxon.GetName())
-        scope = stage.GetPrimAtPath(scope_path)
-        if not scope:
-            scope = stage.DefinePrim(scope_path)
-        if not scope.IsModel():
-            Usd.ModelAPI(scope).SetKind(Kind.Tokens.assembly)
-        path = scope_path.AppendChild(name)
-        if stage.GetPrimAtPath(path):
-            return stage.GetPrimAtPath(path)
-
-        asset_stage = fetch_stage(new_asset_name)
-        asset_origin_path = Sdf.Path("/Origin")
-        asset_origin = asset_stage.DefinePrim(asset_origin_path)
-        category_layer = _find_layer_matching(_TAXONOMY_FIELDS, stage.GetLayerStack())
-        category_layer_id = str(Path(category_layer.realPath).relative_to(repo.get()))
-        asset_origin.GetReferences().AddReference(category_layer_id, taxon.GetPath())
-        asset_stage.SetDefaultPrim(asset_origin)
-
-        if label:
-            label_attr = asset_origin.GetAttribute("label")
-            label_attr.Set(label)
-
-        over_prim = stage.OverridePrim(path)
-        over_prim.GetPayloads().AddPayload(asset_stage.GetRootLayer().identifier)
-
-    root_layer.SetPermissionToEdit(current_permission)
-    return over_prim
+    return create_many(taxon, [name], [label])[0]
 
 
 def taxonomy_context(stage: Usd.Stage) -> Usd.EditContext:
@@ -214,11 +259,12 @@ def taxonomy_context(stage: Usd.Stage) -> Usd.EditContext:
         taxonomy_stage = fetch_stage(taxonomy_asset)
         taxonomy_layer = taxonomy_stage.GetRootLayer()
         # Use paths relative to our repository to guarantee portability
-        taxonomy_reference = str(Path(taxonomy_layer.realPath).relative_to(repo.get()))
-        root_layer.subLayerPaths.append(taxonomy_reference)
+        taxonomy_id = str(Path(taxonomy_layer.realPath).relative_to(Repository.get()))
+        root_layer.subLayerPaths.append(taxonomy_id)
 
         if not taxonomy_stage.GetDefaultPrim():
-            taxonomy_stage.SetDefaultPrim(taxonomy_stage.DefinePrim(_TAXONOMY_ROOT_PATH))
+            default_prim = taxonomy_stage.CreateClassPrim(_TAXONOMY_ROOT_PATH)
+            taxonomy_stage.SetDefaultPrim(default_prim)
 
         return _context(stage, _TAXONOMY_FIELDS)
 
@@ -229,6 +275,12 @@ def unit_context(prim: Usd.Prim) -> Usd.EditContext:
     return _context(prim, fields)
 
 
+def unit_asset(prim: Usd.Prim) -> Sdf.Layer:
+    """Get the asset layer that acts as the 'entry point' for the given prim."""
+    fields = {**_get_id_fields(prim), _UNIT_UNIQUE_ID: prim.GetName()}
+    return _find_layer_matching(fields, _layer_stack(prim))
+
+
 def _root_asset(stage):
     """From a give stage, find the first layer that matches a valid grill identifier.
 
@@ -237,7 +289,7 @@ def _root_asset(stage):
 
     This searches on the root layer first.
     """
-    with suppress(ValueError):
+    with contextlib.suppress(ValueError):
         root_layer = stage.GetRootLayer()
         return UsdAsset(Path(root_layer.identifier).name), root_layer
 
@@ -252,19 +304,20 @@ def _root_asset(stage):
 
 
 def _get_id_fields(prim):
-    data = prim.GetCustomDataByKey(_PRIM_GRILL_KEY) or {}
-    if not data:
-        raise ValueError(f"No data found on '{_PRIM_GRILL_KEY}' key for {prim}")
-    fields = data.get(_PRIM_FIELDS_KEY, {})
+    fields = prim.GetAssetInfoByKey(_ASSETINFO_FIELDS_KEY)
     if not fields:
-        raise ValueError(f"Missing or empty '{_PRIM_FIELDS_KEY}' found on '{_PRIM_GRILL_KEY}' custom data for {prim}. Custom data: {pformat(data)}")
+        raise ValueError(f"Missing or empty '{_FIELDS_KEY}' on '{_ASSETINFO_KEY}' asset info for {prim}. Got: {pformat(prim.GetAssetInfoByKey(_ASSETINFO_KEY))}")
     if not isinstance(fields, typing.Mapping):
-        raise TypeError(f"Expected mapping on key '{_PRIM_FIELDS_KEY}' from {prim} on custom data key '{_PRIM_GRILL_KEY}'. Got instead {fields} with type: {type(fields)}")
+        raise TypeError(f"Expected mapping on key '{_FIELDS_KEY}' from {prim} on custom data key '{_ASSETINFO_KEY}'. Got instead {fields} with type: {type(fields)}")
     return fields
 
 
 def _context(obj, tokens):
-    layers = reversed(list(_layer_stack(obj)))
+    # TODO: do we need to reverse the order of the layer stack?
+    #   at the moment, goes from strongest -> weakest layers
+    if not tokens:
+        raise ValueError(f"Expected a valid populated mapping as 'tokens'. Got instad: '{tokens}'")
+    layers = _layer_stack(obj)
     asset_layer = _find_layer_matching(tokens, layers)
     return _edit_context(obj, asset_layer)
 
@@ -301,8 +354,18 @@ def _(obj: Usd.Stage, layer):
 
 @_edit_context.register
 def _(obj: Usd.Prim, layer):
-    # We need to explicitely construct our edit target since our layer is not on the layer stack of the stage.
-    target = Usd.EditTarget(layer, obj.GetPrimIndex().rootNode.children[0])
+    # We need to explicitly construct our edit target since our layer is not on the layer stack of the stage.
+    # TODO: this is specific about "localised edits" for an asset. Dispatch no longer looking solid?
+    query = Usd.PrimCompositionQuery(obj)
+    query.filter = _ASSET_UNIT_QUERY_FILTER
+    for arc in query.GetCompositionArcs():
+        target_node = arc.GetTargetNode()
+        # contract: we consider the "unit" target node the one matching origin path and the given layer
+        if target_node.path == _UNIT_ORIGIN_PATH and target_node.layerStack.identifier.rootLayer == layer:
+            break
+    else:
+        raise ValueError(f"Could not find appropriate node for edit target for {obj} matching {layer}")
+    target = Usd.EditTarget(layer, target_node)
     return Usd.EditContext(obj.GetStage(), target)
 
 
@@ -318,7 +381,13 @@ def _(obj: Usd.Stage):
 
 @_layer_stack.register
 def _(obj: Usd.Prim):
-    return (spec.layer for spec in obj.GetPrimStack())
+    seen = set()
+    for spec in obj.GetPrimStack():
+        layer = spec.layer
+        if layer in seen:
+            continue
+        seen.add(layer)
+        yield layer
 
 
 class UsdAsset(names.CGAssetFile):
